#ifndef GENOME_H
#define GENOME_H

#include "UnitTest.h"

#include <iostream>
#include <sstream>
#include <map>
#include <string>
#include <vector>
#include <fstream>
#include <ctime>
#include <cstring>
#include <algorithm>
#include <math.h>
#include <pthread.h>

#include "gvector.h"
#include "hash_map.h"
#include "bin_seq.h"
#include "const_include.h"
#include "Reader.h"

/* 
 * The Genome class is used to store the Genome that is loaded in a decent size.
 */

#define NUM_CHAR_VALS	5
#define MAX_NAME_LEN	200
#define MAX_FN_LEN		1024

struct HashLocation {
	HashLocation() {
		size = 0;
		hash_arr = NULL;
	}

	~HashLocation() {
		if(hash_arr)
			delete[] hash_arr;
	}

	unsigned long* hash_arr;
	unsigned long size;
};

class Genome {
	friend class bin_seq;
	
	public:
		/*
		 * Constructor.  Creates new Genome with filename.
		 * fn - the filename of the file to be parsed into the genome.
		 */
		Genome(const char* fn);
		Genome();
		
		~Genome();
		
		void use(const char* fn, unsigned long, unsigned long);
		/**
		 * Wrapper function that just calls the previous one with 0,0
		 */
		void use(const char* fn);
		
		/*
		 * Returns a string from begin to size.  Forward strand of DNA
		 * @par begin - the beginning of the string to return
		 * @par size - the size of the string to return.
		 */
		string GetString(const unsigned long begin, const unsigned int size);
		
		/*
		 * GetMatches will return the vector that represents the given hash.
		 *
		 * @return the vector of the given hash.
		 */
		HashLocation* GetMatches(unsigned int hash);
		
		/**
		 * LoadGenome is responsible for loading the genome into memory--whether it be by reading from
		 * a binary file or from a set of fasta files, it should be the same.
		 * Basically, it does what the hash_and_store function used to do.
		 */
		void LoadGenome();

		/*
		 * hash_and_store() will cause the genome to be hashed--from the file fn given above.
		 * it will then store the genome in a smaller binary array.
		 */
		void hash_and_store();	
		void inc_counter(int,unsigned int&, unsigned long&, unsigned int);
		void ReplaceSpaceWithUnderscore(string &str);

		
		/**
		 * count is used to determine the genome size to split for using MPI
		 */
		unsigned long count();
		
		/*! 
		 * AddScore will add the given score, amt, to the genomic position, pos.
		 *
		 * @param pos - the position on the genome.
		 * @param amt - the amount to be given to this specific position on the genome.
		 */
		float GetScore(const unsigned long &pos);
		void AddScore(const unsigned long &pos, const float &amt);
		void AddScore(const unsigned long &pos, const float &amt, float a_add, float b_add, float gen_a, float gen_b);
		void AddScore(const unsigned long &pos, const float &amt, float a_add, float c_add, float g_add, float t_add,
					float gen_a, float gen_c, float gen_g, float gen_t);

		void AddSeqScore(unsigned long, const float*, const float);
		void AddSeqScore(unsigned long, const float, char which);
		void AddGenScore(unsigned long, const float, char which);
		
		/*! 
		 * GetPos is used mostly in printing out the matches.  It will find the position
		 * on the genome and the corresponding chromosome name.
		 *
		 * @param pos - the position to find the local chromosomal placement from.
		 *
		 * @return a string representing the chromosomal name and local position.
		 */
		string GetPos(const unsigned long &pos, const bool upstrand);
		string GetPos(const pair<unsigned long,bool> p_pos);
		/*!
		 * GetPosPair is used by the ScoredSeq class to return a pair with the sequence name and position--
		 * instead of the single string from GetPos
		 *
		 * @param pos The GNUMAP position of the sequence
		 */
		pair<string,unsigned long> GetPosPair(const unsigned long pos);
		
		/**
		 * Will be used to set the thread ID for this genomic position.
		 * If this location has already been searched by my thread, it will return a -1 and
		 * not set the location.
		 * 
		 * The thread value at pos will always be set to myID after this function is called.
		 */
		//efficiency is not proved, so we're tossing it out
		//int getSetThreadID(long unsigned int pos, pthread_t myID);

		/**
		 * This is needed for the MPI_Reduce call on this genome.  We need to pass back the
		 * pointer to the GenomeLocation object
		 */
		float* GetGenomeAmtPtr() { return amount_genome; }
		float* GetGenomeAPtr() { return reads[0]; }
		float* GetGenomeCPtr() { return reads[1]; }
		float* GetGenomeGPtr() { return reads[2]; }
		float* GetGenomeTPtr() { return reads[3]; }
		float* GetGenomeNPtr() { return reads[4]; }
		
		unsigned long size();
		
		/*!
		 * PrintFinal will print out the entire genome in the SGR format used for viewing in the
		 * Integrated Genome Browser (IGB).  Also prints in SGREX if desired.
		 *
		 * Made a modification, so now it can also print in append mode.  Previous function
		 * is a stub call to the current function.
		 *
		 * @param fn - the filename to read to (will have .sgr appended to the end)
		 * @param append - should I append or write a new file?
		 */
		void PrintFinal(const char* fn, bool append);
		void PrintFinal(const char* fn);
		void AppendFinal(const char* fn);


		inline bool is_max_pos(char,float, float, float, float, float);
		inline unsigned int max_pos(float[], unsigned int);
		/*
		 * is_snp determines if this position is a SNP or not.  It takes as parameters
		 * the character in the genome and the float array represeting the coverage for
		 * all nucleotides at this position.  Currently, there are only two characteristics
		 * that make it a SNP:  It must exceed a pre-defined MIN_COVERAGE amount, and it
		 * must exceed a MIN_RATIO.  
		 * The ratio is the value of the max nucleotide divided by the second highest 
		 * nucleotide
		 * 
		 * @param to_check The character at this position in the genome
		 * @param chars An array representing the nucloetides at all positions
		 * @return -1 if it is not a SNP, else the snp position if it is
		 */
		//inline int is_snp(char, float[5]);
		/**
		 * @param chars An array representing the nucloetides at all positions
		 * @param snp_pos The position for the snp(?)
		 * @return the p-value of the maximum position being the correct nucleotide
		 */
		inline float is_snp(float chars[5], int &snp_pos);
		inline float LRT(float chars[5], int &snp_pos);
		inline float dipLRT(float chars[5], int &snp_pos);

		
		//! Mostly used in testing...
		void print_to_file(hash_map<unsigned int, HashLocation> & gh, char* ofn);

		static bool Test(ostream &os);

	private:
		void Init(const char* fn);
		//unsigned long int store(vector<GenomeLocation> *g, unsigned char* gen, 
		//void store(gvector<GenomeLocation> *g, unsigned char* gen, 
		//						unsigned long int, unsigned long &);
		void store(gvector<GEN_TYPE> *pg, unsigned char* gen, 
								unsigned long int, unsigned long &);
		
		/*!
		 * fix_hash will remove all the high-density matches in the hash.  Anything over the
		 * threshold (MAX_HASH_SIZE) will be removed.
		 */
		void fix_hash();

		inline int base2int(char c);
		inline char int2base(int i);
				
		string reverse(string &s);

		// Reads the genome file from memory
		unsigned int readGen(char* fn);
		// Saves the genome file to memory
		unsigned int saveGen(char* fn);
		
		/**
		 * convertToVector will convert the hashed-and-stored genome to the non-STL version
		 *
		 * @post gh and g will be deleted
		 */
		void convertToVector(hash_map<unsigned int, vector<unsigned long> > *gh,
		                     //gvector<GenomeLocation> *g);
		                     gvector<GEN_TYPE> *pg);

		GEN_TYPE offset;
		unsigned int bin_offset;
		Reader* reader;

		//genome is the compressed genome
		//GenomeLocation* genome;
		GEN_TYPE* packed_genome;
		float* amount_genome;
		
		unsigned long genome_size;

		float* reads[NUM_CHAR_VALS];
		float* gens[NUM_CHAR_VALS];

		//gen_piece is used to store a portion of the genome as we read it in.
		//unsigned char* gen_piece;
		hash_map<unsigned int, HashLocation> gen_hash;
		hash_map<unsigned int, vector<unsigned long> > gen_hash_vec;
		vector<pair<string,unsigned long> > names;
		
		unsigned long my_start;	// The position we should start reading from
		unsigned long my_end;
		bool read_all;
};

#endif
