#ifndef SCORED_SEQ_H
#define SCORED_SEQ_H

#include <string>
#include <set>
#include <cmath>
#include <pthread.h>
#include "const_include.h"
#include "SequenceOperations.h"
#include "GenomeMem.h"

class ScoredSeq {
	protected:
		static const bool usePairHMM = false;

	public:
		string sequence;
		double align_score;
		double log_align_score;
		//the bool in this pair represents which strand:
		//	true = positive strand (+)
		//	false = complementary strand (-)
		set<pair<unsigned long,bool> > positions;

		ScoredSeq() : sequence("<empty>"), align_score(-1),
			log_align_score(exp(align_score)) {
			positions.clear();
		}
	
		ScoredSeq(string seq, double as, unsigned long pos, bool us) :
			align_score(as) {

			if(!us)  {
				//printf("Found a non-rev sequence...changing from %s",seq.c_str());
				sequence = reverse_comp(seq);
				//printf(" to %s\n",sequence.c_str());
			}
			else
				sequence = seq;

			log_align_score = exp(align_score);
			positions.insert(pair<unsigned long,bool>(pos,us));
		}

		ScoredSeq(const ScoredSeq &other) {
			Init(other);
		}
		
		virtual ScoredSeq& operator =(const ScoredSeq &other) {
			//if( !(this == other) )
			Init(other);
			return *this;
		}
		
		virtual ~ScoredSeq() {}
		
		inline void Init(const ScoredSeq &other) {
			positions.clear();
			sequence = other.sequence;
			align_score = other.align_score;
			log_align_score = other.log_align_score;
			positions = other.positions;
		}
		
		inline int compare(ScoredSeq *a, ScoredSeq *b) {
			if(a == b)
				return 0;
			else {
				if(a > b)
					return 1;
				else
					return -1;
			}
		}
		inline bool operator() (ScoredSeq &a, ScoredSeq &b) {
			//printf("ADDR: Comparing %s with %s\n",a.sequence.c_str(),b.sequence.c_str());
			return a.sequence.compare(b.sequence);
		}
		
		//The return value is 1 if the set doesn't contain an element whose sort key matches the parameter key. 
		//0 if the set does contain an element with a matching key.
		inline bool operator() (const ScoredSeq *a, const ScoredSeq *b) const {
			//printf("Comparing %s with %s\n",a->sequence.c_str(),b->sequence.c_str());
			return (a->sequence != b->sequence);
		}
		inline bool operator() (const double &a, const float &b) {
			return a != b;
		}
		inline bool operator == (ScoredSeq &other) {
			if( (sequence == other.sequence) )
				return true;
			return false;
		}
		inline bool operator == (ScoredSeq *other) {
			if( (sequence == other->sequence) )
				return true;
			return false;
		}
		inline bool operator< (const ScoredSeq &other) const {
			//printf("< adr Comparing %s with %s\n",sequence.c_str(),other.sequence.c_str());
			//return align_score < other.align_score;
			return (sequence < other.sequence);
		}
		inline bool operator< (const ScoredSeq* other) const {
			//printf("< ptr Comparing %s with %s\n",sequence.c_str(),other->sequence.c_str());
			//return align_score < other->align_score;
			return (sequence < other->sequence);
		}
		inline bool operator> (const ScoredSeq &other) const {
			//printf("> adr Comparing %s with %s\n",sequence.c_str(),other.sequence.c_str());
			double my_total = log_align_score / positions.size();
			double other_total = other.log_align_score / other.positions.size();
			return my_total > other_total;
		}
		inline bool operator> (const ScoredSeq *other) const {
			//printf("> ptr Comparing %s with %s\n",sequence.c_str(),other->sequence.c_str());
			double my_total = log_align_score / positions.size();
			double other_total = other->log_align_score / other->positions.size();
			return my_total > other_total;
		}
		
		inline bool is_greater(const ScoredSeq &other) {
			//printf("Logs: %f=%e %f=%e\n",other.align_score,exp(other.align_score),align_score,exp(align_score));
			double my_total = log_align_score;
			double other_total = other.log_align_score;
			return my_total > other_total;
		}
		
		double get_score() {
			return align_score;
		}
		double get_log_align_score() {
			return log_align_score;
		}
		
		bool add_spot(unsigned long pos, bool us) {
			//cout << "inserting: " << pos << endl;
			if( positions.insert( pair<unsigned long,bool>(pos,us) ).second ) {
				//cout << "inserted. \tsize: " << positions.size() << endl;
				return true;
			}
			else {
				//cout << "already in set. \tsize: " << positions.size() << endl;
				return false;
			}
		}
		
		/*
		 * print_full_w_score will print the full score of this sequence, returning
		 * the sequence information with its output.
		 * It will NOT modify the Genome object.
		 * 
		 * @return A string representing the Sequence
		 */
		virtual string print_full(double denom, Genome &gen, Read &read) const {
			ostringstream temp;
			double total_score = log_align_score / denom;
			temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << sequence;
				
			set<pair<unsigned long,bool> >::iterator sit;
			
			for(sit = positions.begin(); sit != positions.end(); sit++)
				temp << "\t" << gen.GetPos(*sit);
				
			return temp.str();
		}
		
		/*
		 * print_and_score will print the full score of this sequence, returning the
		 * sequence information for its output.
		 * This function WILL add its value to the Genome object
		 * 
		 * @return A string representing the Sequence
		 */
		virtual string print_and_score(double denom, Genome &gen) const {
			ostringstream temp;
			double total_score = log_align_score / denom;
			temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << sequence;
			
			set<pair<unsigned long,bool> >::iterator sit;
			
			for(sit = positions.begin(); sit != positions.end(); sit++) {
				temp << "\t" << gen.GetPos(*sit);
				gen.AddScore( (*sit).first,total_score );
			}
				
			return temp.str();
		}
		
		/*
		 * score will add the log-odds score (adjusted by denom) to each
		 * position on the Genome gen.  Additionally, this can be passed
		 * a flag to specify the number of positions after the start position
		 * to add to (for adding to different lengths of sequences).
		 */		 
		virtual void score(double denom, Genome &gen, unsigned int len, Read &read, pthread_mutex_t &lock) = 0;
		 
/*		 
		virtual void score(double denom, Genome &gen, unsigned int len, Read &read) const {

			double total_score = log_align_score / denom;
			set<pair<unsigned long,bool> >::iterator sit;
				
			for(sit=positions.begin(); sit!=positions.end(); sit++)
				for(unsigned int i=0; i<len; i++) 
					gen.AddScore( (*sit).first+i,total_score );
		
		}
*/
	
		// We'll just put this here until we deal with it later
		/*
		//void score_max_atog(double denom, Genome &gen, unsigned int len, vector<vector<double> > &pwm, ostringstream &str) const {
		void score_max_atog(double denom, Genome &gen, unsigned int len, Read &read, ostringstream &str) const {
			if(!positions.size())
				return;
			double total_score = log_align_score / denom;
			//set<pair<unsigned int,bool> >::iterator sit;
		
			bin_seq bs;	//use the global one
			unsigned int seq_size = sequence.size();
	
			if(!positions.size())	//if there were no matches, just return.
				return;
	
			string read_consensus = GetConsensus(read);
	
			string gen_string = gen.GetString((*positions.begin()).first, seq_size);
			string aligned = bs.get_align_score_w_traceback(read,read_consensus,gen_string);
	
			str << aligned << "\t";
	
			set<pair<unsigned long,bool> >::iterator sit;
			for(sit=positions.begin(); sit!=positions.end(); sit++) {
				//string gen_string = gen.GetString((*sit).first, seq_size);
				//string aligned = bs.get_align_score_w_traceback(read, sequence, gen_string);
				//if(!aligned.equals(prev_aligned))
				//	printf("Alignments are Not Equal!\n");
				
				for(unsigned int i=0; i<aligned.size(); i++) {
					if(gen_string[i] == 'a') {
						gen.AddGenScore((*sit).first+i,total_score,'a');
						switch(aligned[i]) {
							case 'a':
								gen.AddSeqScore((*sit).first+i,total_score,'a');
							break;
							case 'c':
								gen.AddSeqScore((*sit).first+i,total_score,'c');
							break;
							case 'g':
								gen.AddSeqScore((*sit).first+i,total_score,'g');
							break;
							case 't':
								gen.AddSeqScore((*sit).first+i,total_score,'t');
							break;
						}
					}
				}
			}
		}
		*/
		
			/**
			 * Since the SAM format took just a little while for me to understand, I'll explain it
			 * here.  According to specs, each sequence is as follows:
			 * 
			 * <QNAME> <FLAG> <RNAME> <POS> <MAPQ> <CIGAR> <MRNM> <MPOS> <ISIZE> <SEQ> <QUAL>
			 * 
			 * - For GNUMAP, the QNAME will be used if it is a fastq or fasta string--otherwise
			 *   the query name will be "seq" followed by the sequence number in the file.
			 * - The only bits in FLAG that are important are 0x0010 (where 1 represents an alignment
			 *   to the forward strand and 0 represents reverse)
			 * - RNAME is the chromosome mapped to
			 * - POS is the position on the chromosome
			 * - MAPQ is the mapping quality, where MAPQ = 10 * log_10(1-p(x)) where p(x) is GNUMAP's
			 *   posterior probability of mapping to that specific location
			 * - CIGAR is the alignment differences, where I=>Insertion, D=>Deletion and M=>Mismatch
			 *   on the genomic strand
			 * - MRNM will be '=' for all sequences (it's the mate-pair name)
			 * - MPOS is ignored.  Always '0'
			 * - ISIZE is ignored.  Always '0'
			 * - SEQ is the sequence
			 * - QUAL is the Phred-based quality of the sequences
			 * 
			 * @param denom The denominator for this particular read
			 * @param read The GNUMAP-style Read for this sequence
			 * @param consensus The consensus sequence for this Read
			 * @param gen The Genome that has been used for this program
			 * 
			 * @return A string in the SAM format
			 */
			string print_SAM(double denom, Read &read, string consensus, Genome &gen) {
				ostringstream to_return;
				bin_seq bs;
				
				double total_score = log_align_score / denom;
				
				// Mapping quality is defined as -10 * log_10(1-p)
				int MAPQ = (int)round(-10 * log(total_score)/log(10));

				string QUAL = str2qual(read);
				//fprintf(stderr,"QUAL for %s is %s\n",read.name,QUAL.c_str());

				if(positions.size() < 1) {

					//return "";

					to_return << read.name << "\t";
					
					to_return << 0x0000 << "\t";
					to_return << "*\t0\t";
					to_return << "0" << "\t";
					to_return << "*\t";
					to_return << "=\t0\t0\t";
					to_return << consensus << "\t";
					to_return << QUAL.c_str() << "\n";
					return to_return.str();
				}

				string gen_string = sequence;
				//if((*positions.begin()).second == false)
				//	gen_string = reverse_comp(sequence);

				pair<string,string> aligned = bs.get_align_score_w_traceback(read,consensus,gen_string);

				string CIGAR = aligned.second;
				if(CIGAR.length() == 0)
					CIGAR = "*";

				// Print all the locations
				set<pair<unsigned long, bool> >::iterator sit;
				for(sit = positions.begin(); sit != positions.end(); sit++) {
					to_return << read.name << "\t";

					if((*sit).second)
						to_return << 0x0010 << "\t";
					else
						to_return << 0x0000 << "\t";
					
					pair<string,unsigned long> seq_pos = gen.GetPosPair((*sit).first);
					to_return << seq_pos.first << "\t" << seq_pos.second << "\t";
					//to_return << "(" << total_score << ")" << MAPQ << "\t";
					to_return << MAPQ << "\t";
					to_return << CIGAR << "\t";
					to_return << "=\t0\t0\t";
					to_return << consensus << "\t";
					to_return << QUAL.c_str() << "\n";
				}

				return to_return.str();
			}
			
			vector<TopReadOutput> get_SAM(double denom, Read &read, string consensus, Genome &gen, unsigned int rIndex) {
				bin_seq bs;

				vector<TopReadOutput> ret_vec;
				if(positions.size() == 0)
					return ret_vec;
					
				double total_score = log_align_score / denom;
				
				// Mapping quality is defined as -10 * log_10(1-p)
				int MAPQ = (int)round(-10 * log(total_score)/log(10));

				string gen_string = sequence;
				//if((*positions.begin()).second == false)
				//	gen_string = reverse_comp(sequence);

				pair<string,string> aligned = bs.get_align_score_w_traceback(read,consensus,gen_string);

				string CIGAR = aligned.second;
				if(CIGAR.length() == 0)
					CIGAR = "*";

				// Print all the locations
				set<pair<unsigned long, bool> >::iterator sit;
				for(sit = positions.begin(); sit != positions.end(); sit++) {
					pair<string,unsigned long> seq_pos = gen.GetPosPair((*sit).first);

					TopReadOutput tro;
					strncpy(tro.READ_NAME,read.name,100-1);
					tro.READ_NAME[100-1] = '\0';
					strncpy(tro.CHR_NAME,seq_pos.first.c_str(),100-1);
					tro.CHR_NAME[100-1] = '\0';
					tro.CHR_POS = seq_pos.second;
					tro.strand = (*sit).second;
					tro.MAPQ = MAPQ;
					strncpy(tro.CIGAR,CIGAR.c_str(),100-1);
					tro.CIGAR[100-1] = '\0';
					tro.readIndex = rIndex;
					
					ret_vec.push_back(tro);
				}

				return ret_vec;
			}

};

#endif
