#ifndef _SEQMANAGER_H
#define _SEQMANAGER_H

#include <vector>
#include <string>
#include <pthread.h>
#include <cassert>
#include "const_include.h"
#include "SeqReader.h"

#define READS_PER_PROC	1024

class SeqManager {

	public:
		SeqManager(Read** reads, const char* readsStr, const unsigned int nR, const unsigned int nP,
			const unsigned int nN, const unsigned int iproc) 
			: g_nReads(nR), g_nProc(nP), g_nNodes(nN), g_iproc(iproc) {
				
			Init(reads, readsStr, nR, nP);
		}
		
		SeqManager(Read** reads, const char* readsStr, const unsigned int nR, const unsigned int nP)
			: g_nReads(nR), g_nProc(nP), g_nNodes(0), g_iproc(0) {

			Init(reads, readsStr, nR, nP);		
		}
		
		~SeqManager() {
			if(sr)
				delete sr;
		}
		
		void Init(Read** reads, const char* readsStr, const unsigned int nR, const unsigned int nP) {
			finished = false;

			g_readPtr = reads;
			
			sr = new SeqReader(readsStr);
		
			//if(gVERBOSE) 
				fprintf(stderr,"Matching %d file(s):\n",sr->GetNumSeqFiles());
		
			readFiles = sr->GetSeqFileVector();
			rit = readFiles.begin();
			getNextFile();

			totalSeqs = 0;
			readCount = 0;
			unsigned int rpp = g_nReads/g_nProc;
			//readsPerProc = rpp > READS_PER_PROC ? READS_PER_PROC : rpp;
			readsPerProc = rpp;

			// init the read lock
			pthread_mutex_init(&read_lock,NULL);

		}
		
		/**
		 * Resets the counter--adds to the beginning of the global read array
		 */
		void resetCounter() {
			if(gVERBOSE)
				printProgress();
			readCount = 0;
		}
		
		unsigned int getTotalSeqCount() {
			return totalSeqs;
		}

		void printProgress() {
			fprintf(stderr,"[%d/-] %.0f%% reads complete\n",g_iproc,((float)fileReadCount)/readPerFile * 100);
		}
		
		unsigned int getCurrNumReads() {
			return readPerFile;
		}

		/**
		 * Here, we'll read all the reads we can into an array passed in.  This allows the mother
		 * processor to decide how to divy them out.
		 * 
		 * Used by OpenMP
		 * 
		 * @param readArr The read pointer array to which we should add the reads
		 * @param size The size of the read pointer array (readArr)
		 * @param num_reads Will be the number of reads we've added to the array
		 * @return true if we added any reads, false if we're entirely done adding reads
		 */	
		bool fillReadArray(Read** readArr, const unsigned int size, unsigned int &num_reads) {
			unsigned int i;
			Read* temp;

			pthread_mutex_lock(&read_lock);

			num_reads = 0;

			if(finished) {
				// Unlock the mutex first, then return
				pthread_mutex_unlock(&read_lock);
				return false;
			}

			for(i=0; i<size; i++) {
				try {
					while((temp = sr->GetNextSequence()) == NULL) {
						if(!getNextFile()) {
							//fprintf(stderr,"CANNOT READ MORE FROM FILE...RETURNING\n");
							pthread_mutex_unlock(&read_lock);
							return true;
						}
					}
				}
				catch(Exception* e) {
					fprintf(stderr,"ERROR:  %s\n",e->GetMessage());
					throw e;
				}
				

				// Set the pointer
				readArr[i] = temp;
				num_reads++;
			}

			fileReadCount += num_reads;
			
			pthread_mutex_unlock(&read_lock);

			return true;
		}
	
		
		/**
		 * Responsible for dishing out reads, as needed
		 * Parameter of 'set' added to be backwards compatable
		 * 
		 * @par begin The beginning location of where the reads will be placed
		 * @par end The ending location of where the reads will be placed
		 * @par set true iff we should set 'begin', false if we should use the value of 'begin'
		 */
		bool getMoreReads(unsigned int &begin, unsigned int &end, bool set) {
			// Ensure there's only one thread in here
			pthread_mutex_lock(&read_lock);
			
			
			if(finished) {
				//fprintf(stderr,"[%d/-] CANNOT READ MORE--RETURNING\n",g_iproc);
				if(set)
					begin = 0; end = 0;
				pthread_mutex_unlock(&read_lock);
				return false;
			}
			
			
				
			Read* temp;
			unsigned int i;
			unsigned int num_reads=0;
			bool ret_val = true;
			
			if(set)
				begin = readCount;

			//fprintf(stderr,"[%d/-] Starting with begin=%u,end=%u and fileProgress=%d/%d\n",g_iproc,begin,end,fileReadCount,readPerFile);

			for(i=0; i< (set ? readsPerProc : end-begin) && fileReadCount+i<readPerFile; i++) {
				// if there's no more reads, get the next sequence from the next file
				while((temp = sr->GetNextSequence()) == NULL) {
					//fprintf(stderr,"[%d/-] temp is NULL (i=%d)\n",g_iproc,i);
					if(!getNextFile()) {
						//fprintf(stderr,"[%d/-] CANNOT READ MORE FROM FILE...RETURNING (%u-%u) and finished is %d\n",g_iproc,begin,end,finished);
						//fprintf(stderr,"[%d/-] i:%u, total=%u\n",g_iproc,i,fileReadCount);
						ret_val = false;
						goto AFTER_FOR;
					}
				}
				//fprintf(stderr,"[%d/-] temp is NOT NULL!!!\n");

					
				// Set it at the global pointer
				g_readPtr[begin+i] = temp;
				num_reads++;
			}
AFTER_FOR:			
			
			if(set) {
				end = begin+num_reads;
				readCount = end;
			}

			//fprintf(stderr,"Read %d reads, start is %d and end is %d\n",num_reads,begin,end);
			totalSeqs += num_reads;
			fileReadCount += num_reads;
			
			if(fileReadCount >= readPerFile) {
				finished = true;
				ret_val = false;
			}

		
			//fprintf(stderr,"Read from %u to %u and has read %u(%u) seqs\n",begin,set ? readsPerProc : end-begin,num_reads,totalSeqs);
			// Release the lock
			pthread_mutex_unlock(&read_lock);
			
			return ret_val;
		}
		
		/**
		 * Responsible for dishing out reads, as needed.  Used with previous version of GNUMAP
		 */
		/*
		bool getMoreReads(Read* reads[], const unsigned int &length) {
			if(finished) {
				return false;
			}
			
			// Ensure there's only one thread in here
			pthread_mutex_lock(&read_lock);
				
			Read* temp;
			unsigned int i;
			unsigned int num_reads=0;
			
			for(i=0; i<readsPerProc; i++) {
				// if there's no more reads, get the next sequence from the next file
				while((temp = sr->GetNextSequence()) == NULL) {
					if(!getNextFile()) {
						//fprintf(stderr,"CANNOT READ MORE FROM FILE...RETURNING\n");
						goto AFTER_FOR;
					}
				}

					
				// Set it at the global pointer
				reads[num_reads] = temp;
				num_reads++;
			}
//AFTER_FOR:			
			
			//fprintf(stderr,"Read %d reads, start is %d and end is %d\n",num_reads,begin,end);
			totalSeqs += num_reads;
			fileReadCount += num_reads;

		
			// Release the lock
			pthread_mutex_unlock(&read_lock);
			
			return true;
		}

		unsigned int getNumSeqsPProc() {
			//return readPerFile/g_nProc;
			return readsPerProc;
		}
		*/
		
		bool isFinished() {
			return finished;
		}
		
	private:
	
		bool getNextFile() {
			
			bool invalid = true;
			while(invalid) {
				if(rit == readFiles.end()) {
					finished = true;
					return false;
				}

				try{

					unsigned int nBurn = 0;
					unsigned int total_reads = sr->GetNumSeqs(*rit);
				
					// Need to burn sequences to be at the right pointer for each node
					if(g_nNodes) {
						// use +1 because we want to get all of them
						unsigned int each = total_reads/g_nNodes+1;
						unsigned int begin = each * g_iproc;
						
						nBurn = begin;
				
						readPerFile = each;
					}					
					else {
						readPerFile = total_reads;
					}

					//fprintf(stderr,"[%d/-] Should burn %u reads and have %u reads total. total_reads:%u\n",g_iproc,nBurn,readPerFile,total_reads);					
					// instruct the SeqReader to use this file
					sr->use(*rit,nBurn);
					
					// increment the vector iterator
					rit++;

					invalid = false;
					fileReadCount = 0;
				}
				catch(const char* err) {
					fprintf(stderr,"ERROR: \n\t%s\n",err);
					fprintf(stderr,"\tIn file: %s\n",(*rit).c_str());
					continue;
				}
				catch(Exception *e) {
					fprintf(stderr,"ERROR: \n\t%s\n",e->GetMessage());
					fprintf(stderr,"\tIn file %s\n",(*rit).c_str());
					delete e;
					continue;
				}
				catch(...) {
					fprintf(stderr,"ERROR: \n\tImproper sequence length or file.\n");
					fprintf(stderr,"\tIn file %s\n",(*rit).c_str());
					continue;
				}
			}
			
			
			if(gVERBOSE)
				fprintf(stderr,"\n[%d/-] Matching %u sequences of: %s\n",g_iproc, readPerFile,sr->GetFilename().c_str());
						
			return true;
		}
		
		Read** g_readPtr;			// Read pointer from Driver
		SeqReader *sr;				// SeqReader to get Reads from
		vector<string> readFiles;	// The vector from the parsed read files
		vector<string>::iterator rit;	// Vector Iterator for the read files
		unsigned int readPerFile;	// Number of reads per file
		unsigned int readCount;		// Location in the global array for the Read to go
		unsigned int fileReadCount;	// For the percentage of reads that are read
		unsigned int totalSeqs;
		unsigned int g_nReads;		// Number of reads in the global array
		unsigned int g_nProc;		// Number of processors in the program
		unsigned int g_nNodes;		// Number of nodes (MPI) in the program
		unsigned int g_iproc;		// My processor number (MPI)
		unsigned int readsPerProc;	// Number of reads per processor=g_nReads/g_nProc
		
		bool finished;				// Flag when we're finished
		
		pthread_mutex_t read_lock;	// Lock to ensure no duplication
};

#endif // _SEQ_MANAGER_H
