#ifndef BIN_SEQ_H
#define BIN_SEQ_H

#include "UnitTest.h"
#include "const_include.h"
#include "Reader.h"
#include <cassert>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>

using namespace std;

#define DEF_ARR_SIZE	50*50


enum CIGAR_TYPE {
	MATCH, INS, DEL
};

#define CIGAR_TO_STR(k) (k == MATCH) ? 'M' : (k == INS) ? 'I' : 'D'


class bin_seq {
	private:
		// Pair HMM Values
		float PHMM_q;
		float PHMM_t;
		float PHMM_d;
		float PHMM_e;
		float PHMM_Tmm;
		float PHMM_Tgm;
		float PHMM_Tmg;
		float PHMM_Tgg;


	public:
		bin_seq() {
			PHMM_q = 0.25;
			PHMM_t = 0.05;
			PHMM_d = 0.1;
			PHMM_e = 0.5;
			PHMM_Tmm = 1-2*PHMM_d-PHMM_t;
			PHMM_Tgm = 1-PHMM_d-PHMM_t;
			PHMM_Tmg = PHMM_d;
			PHMM_Tgg = PHMM_e;
		
			def_arr = new float[DEF_ARR_SIZE];
			def_arr_size = DEF_ARR_SIZE;
			def_moves_arr = 0;
			def_moves_size = 0;
		};

		~bin_seq() {
			delete[] def_arr;
			def_arr_size = 0;
			if(def_moves_arr)
				delete[] def_moves_arr;
			def_moves_size = 0;
		}
		
		unsigned int str2int();
		unsigned int str2int(string &a);
		string hash2str(const unsigned int h, int length);
		
		/*
		 * get_hash will get the hashing value for either a given string or
		 * a given location on genome.  The return value of get_hash is a pair
		 * of bool and BIT64_t, representing whether the hash was successful, and,
		 * if it was, the value for the hash.
		 * get_hash will return false only if there is an unknown character found in
		 * the middle of the sequence.  For example, actgNgcta will not be a unique 
		 * 9-mer, so it is not added to the hash.  In addition, when trying to map
		 * back to a genome, if the read has an unknown character in the given K-mer,
		 * it should not map back exactly to any portion of the genome.
		 */
		pair<bool,unsigned long> get_hash(string str);
		pair<bool,unsigned long> get_hash(const char* str);
		pair<int,unsigned long> get_hash(unsigned int &pos, const unsigned char* genome);
		pair<int,unsigned long> get_hash(unsigned int &pos, const unsigned char* genome, const unsigned int max_pos);
		
		inline unsigned int base2int(char c);
		inline char int2base(unsigned int i);
		
		float get_align_score(const Read&, const string&, unsigned int, unsigned int);
		float get_align_score_begin(const Read &, const string &, unsigned int);
		float get_align_score_mid(const Read &, const string &, unsigned int, unsigned int);
		float get_align_score_end(const Read &, const string &, unsigned int);
		float get_align_score(Read &read, string &gen);

		/*
		 * pairHMM will perform the pair HMM on the two strings, the one being the read and
		 * the other the genome.  The result will come in a double array of floats representing
		 * the amount at each position.  This can then be mapped to the genome.
		 */
		float** pairHMM(const Read&, const string&, const string&);
		inline float p(char, char);
		inline float p_seq(float x[4], char);
		
		//string get_align_score_w_traceback(const Read &read, const string &consense, const string &gen);
		/**
		 * The paired version of the get_align_score_w_traceback will align the two sequences and
		 * return strings that represent the alignment.  The first string will be the result of the
		 * alignment in the first sequence.  The second sequence will be the SAM-style CIGAR 
		 * alignment values, where only the following characters are used:
		 * 
		 * M => Alignment Match (can be match or mismatch)
		 * I => Insertion into Reference
		 * D => Deletion from the Reference
		 *
		 * Unlike the SAM output format where the resulting sequence does not have any gaps, the
		 * first sequence will be the gapped sequence so the .sgrex format can still be used.
		 *
		 * @param read - The pwm representing the read to be aligned
		 * @param consense - The consensus sequence of the read
		 * @param gen - The genomic sequence to which this sequence will be mapped
		 * 
		 * @return pair<string,string> where pair.first is the alignment sequence (with dashes for
		 *		gaps) and pair.second is the CIGAR of the alignment
		 */
		pair<string,string> get_align_score_w_traceback(const Read &read, const string &consense, const string &gen);
		int get_align_score(string seq1, string seq2);	//archaic
		
		static bool Test(ostream &os);
		
		
	private:	
		float get_val(char a, char b);
		float get_val(const float* a, const char b);

		int max(int a, int b, int c);
		//double max_dbl(double a, double b, double c);
		float max_flt(float a, float b, float c);
		unsigned int max_flt(float array[], unsigned int array_size);
		float max_flt(char&,float, float, float);
		void print_align(string str1, string align, string str2);

		float* def_arr;
		unsigned int def_arr_size;
		char* def_moves_arr;
		unsigned int def_moves_size;
		
};

#endif
